/***************************************************************************
 *   Copyright (C) 2007 by Saritha Kalyanam   				   *
 *   sari.kaly@gmail.com   						   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#define _GNU_SOURCE
#include <getopt.h>

#include <errno.h>
extern int errno;

#include <net/if.h>
#include <ifaddrs.h>
#include <net/if.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <poll.h>

#include "tinydht.h"
#include "dht_types.h"
#include "dht.h"
#include "pkt.h"
#include "debug.h"
#include "types.h"
#include "crypto.h"
#include "stun.h"
#include "queue.h"
#include "float.h"

extern int h_errno;

/*--------------- Global Variables -----------------*/

char rpc_ifname[IFNAMSIZ];
int n_rpc_ifs = 0;
struct dht_net_if rpc_if[MAX_DHT_NET_IF];

char svc_ifname[IFNAMSIZ];
int n_svc_ifs = 0;
struct dht_net_if svc_if[MAX_DHT_NET_IF];

int n_dhts = 0;
struct dht *dht[MAX_DHT_INSTANCE];

TAILQ_HEAD(task_list_head, task) task_list;

int n_poll_fds = 0;
int poll_fd[MAX_POLL_FD];

/*--------------- Private Functions -----------------*/

int tinydht_init(void);
void tinydht_exit(void);

int tinydht_get_intf_ip_addrs(const char *ifname, 
                            struct dht_net_if *nif, int *n_if, int max_if);
int tinydht_get_intf_ext_ip_addr(struct dht_net_if *nif);

int tinydht_init_service(void);

int tinydht_add_dht(unsigned int type, struct dht_net_if *nif);

int tinydht_poll_loop(void);
int tinydht_task_schedule(void);

bool tinydht_is_service_fd(int fd);
struct dht * tinydht_find_dht_from_fd(int fd);

int tinydht_decode_request(int sock, struct sockaddr_storage *from, 
                            size_t fromlen, u8 *data, int len);
int tinydht_put(struct tinydht_msg *msg);
int tinydht_get(struct tinydht_msg *msg);

/*--------------- Implementation -----------------*/

int
main(int argc, char *argv[])
{
    int ret;

    memcpy(rpc_ifname, "eth0", sizeof(rpc_ifname));
    memcpy(svc_ifname, "eth0", sizeof(svc_ifname));

    ret = tinydht_init();
    if (ret != SUCCESS) {
        ERROR("TinyDHT initialization failed!\n");
        return EXIT_FAILURE;
    }

    /* the main poll loop */
    tinydht_poll_loop();

    /* we should never be here!
     * but if we do, cleanup */
    tinydht_exit();

    return EXIT_SUCCESS;
}

int
tinydht_init(void)
{
    int ret;
    int i;
    
    /* initialize the crypto engine */
    ret = crypto_init();
    if (ret != SUCCESS) {
        return ret;
    }

    /* get local ip addr */
    ret = tinydht_get_intf_ip_addrs(rpc_ifname, rpc_if, 
                                        &n_rpc_ifs, MAX_DHT_NET_IF); 
    if (ret != SUCCESS) {
        return EXIT_FAILURE;
    }

    ret = tinydht_get_intf_ip_addrs(svc_ifname, svc_if, 
                                        &n_svc_ifs, MAX_DHT_NET_IF); 
    if (ret != SUCCESS) {
        return EXIT_FAILURE;
    }

    for (i = 0; i < n_rpc_ifs; i++) {
        ret = tinydht_get_intf_ext_ip_addr(&rpc_if[i]);
        if (ret != SUCCESS) {
            /* FIXME: should we abort here? */
            continue;
        }
    }

    /* initialize the task list */
    TAILQ_INIT(&task_list);
    
    /* instantiate a dht for each rpc intf */
    for (i = 0; i < n_rpc_ifs; i++) {
        ret = tinydht_add_dht(DHT_TYPE_AZUREUS, &rpc_if[i]);
        if (ret != SUCCESS) {
            return EXIT_FAILURE;
        }
    }

    /* initialize the tinydht service */
    ret = tinydht_init_service();
    if (ret != SUCCESS) {
        return EXIT_FAILURE;
    }

    return SUCCESS;
}

void
tinydht_exit(void)
{
}

int
tinydht_get_rnd_port(u16 *port)
{
    int ret;
    u16 s;

    ret = crypto_get_rnd_short(&s);
    if (ret != SUCCESS) {
        return ret;
    }

    *port = htons(1024 + s % (65536 - 1024));
    
    return SUCCESS;
}

int
tinydht_get_intf_ip_addrs(const char *ifname, 
                            struct dht_net_if *nif, int *n_if, int max_if)
{
    struct ifaddrs *ifap = NULL, *ifa = NULL;
    bool found = FALSE;
    int ret;

    if (getifaddrs(&ifap) != 0) {
        return FAILURE;
    }

    for (ifa = ifap; ifa != NULL; ifa = ifa->ifa_next) {

        if (*n_if >= max_if) {
            break;
        }

        if (strncmp(ifa->ifa_name, ifname, IFNAMSIZ) != 0)
            continue;

        if (!(ifa->ifa_addr != NULL))
            continue;

        if (!(ifa->ifa_flags & IFF_UP))
            continue;

        switch (ifa->ifa_addr->sa_family) {
            case AF_INET:
                found = TRUE;
                ret = dht_net_if_new(&nif[*n_if], ifname, 
                                    (struct sockaddr *)ifa->ifa_addr, 
                                    sizeof(struct sockaddr_in));
                if (ret != SUCCESS) {
                    goto err;
                }
                (*n_if)++;
                break;

            case AF_INET6:
                found = TRUE;
                ret = dht_net_if_new(&nif[*n_if], ifname, 
                                    (struct sockaddr *)ifa->ifa_addr, 
                                    sizeof(struct sockaddr_in6));
                if (ret != SUCCESS) {
                    goto err;
                }
                (*n_if)++;
                break;

            default:
                break;
        }
    }

    freeifaddrs(ifap);
    
    if (!found) {
        ERROR("%s - not a working intf\n", ifname);
        return FAILURE;
    }

    return SUCCESS;

err:
    freeifaddrs(ifap);
    
    return FAILURE;
}

int
tinydht_get_intf_ext_ip_addr(struct dht_net_if *nif)
{
    int sa_family;
    struct sockaddr_in addr4, ext_addr4;
    int ret;

    sa_family = ((struct sockaddr *)&nif->local_addr)->sa_family;

    switch (sa_family) {
        case AF_INET:
            bzero(&addr4, sizeof(struct sockaddr_in));
            memcpy(&addr4, &nif->local_addr, sizeof(struct sockaddr_in));

            bzero(&ext_addr4, sizeof(struct sockaddr_in));

            ret = tinydht_get_rnd_port((u16 *)&addr4.sin_port);
            if (ret != SUCCESS) {
                return ret;
            }

            printf("our STUN port %d\n", ntohs(addr4.sin_port));

            /* use STUN to find out the external address */
            ret = stun_get_ext_ip_addr(&addr4, &ext_addr4);
            if (ret != SUCCESS) {
                return ret;
            }

            INFO("TinyDHT RPC Public IP %s\n", inet_ntoa(ext_addr4.sin_addr));

            memcpy(&nif->ext_addr, &ext_addr4, sizeof(struct sockaddr_in));
            nif->ext_addr.ss_family = AF_INET;

            break;


        case AF_INET6:
            /* FIXME: no STUN support for IPv6? */
            goto err;

        default:
            goto err;
    }
   
    return SUCCESS;

err:
    return FAILURE;
}

int
tinydht_add_dht(unsigned int type, struct dht_net_if *nif)
{
    struct dht *d = NULL;
    unsigned short port;
    int ret = FAILURE;
    int i;
    bool unique_port;

    do {
        unique_port = TRUE;
        ret = tinydht_get_rnd_port((u16 *)&port);
        if (ret != SUCCESS) {
            return ret;
        }

        for (i = 0; i < n_dhts; i++) {
            /* don't pick the tinydht svc port */
            if (port == TINYDHT_SERVICE) {
                unique_port = FALSE;
                break;
            }
       
            /* don't pick some other dht instance's port */
            if (dht[i]->port == port) {
                unique_port = FALSE;
                break;
            }
        }

    } while (!unique_port);

    for (i = 0; (i < MAX_DHT_TYPE) && dht_table[i]; i++) {
        if (dht_table[i]->type == type) {
            d = dht_table[i]->constructor(nif, htons(port));
            if (!d) {
                /* FIXME: should we abort here? */
                ret = SUCCESS;
                break;
            }
            dht[n_dhts] = d;
            n_dhts++;
            ret = SUCCESS;
            break;
        }
    }

    return ret;
}

int
tinydht_init_service(void)
{
    int i;
    int sock;
    int ret;
    struct sockaddr_in addr4;
    struct sockaddr_in6 addr6;

    for (i = 0; i < n_svc_ifs; i++) {

        switch (svc_if[i].local_addr.ss_family) {
            case AF_INET:
                sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
                if (sock < 0) {
                    /* FIXME: should we abort here? */
                    continue;
                }
                bzero(&addr4, sizeof(struct sockaddr_in));
                memcpy(&addr4, &svc_if[i].local_addr, 
                        sizeof(struct sockaddr_in));
                addr4.sin_port = htons(TINYDHT_SERVICE);
                ret = bind(sock, (struct sockaddr *)&addr4, 
                            sizeof(struct sockaddr_in));
                if (ret < 0) {
                    ERROR("bind() - %s\n", strerror(errno));
                    break;
                }

                ret = tinydht_add_poll_fd(sock);
                if (ret != SUCCESS) {
                    break;
                }

                ret = listen(sock, 10);
                if (ret != SUCCESS) {
                    ERROR("listen() - %s\n", strerror(errno));
                    break;
                }

                svc_if[i].sock = sock;

                INFO("TinyDHT IPv4 service listening on port %hu\n", 
                        TINYDHT_SERVICE);

                break;

            case AF_INET6:
                sock = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
                if (sock < 0) {
                    /* FIXME: should we abort here? */
                    continue;
                }
                bzero(&addr6, sizeof(struct sockaddr_in6));
                memcpy(&addr6, &svc_if[i].local_addr, 
                        sizeof(struct sockaddr_in6));
                addr6.sin6_port = htons(TINYDHT_SERVICE);
                ret = bind(sock, (struct sockaddr *)&addr6, 
                            sizeof(struct sockaddr_in6));
                if (ret < 0) {
                    ERROR("bind() - %s\n", strerror(errno));
                    break;
                }

                ret = tinydht_add_poll_fd(sock);
                if (ret != SUCCESS) {
                    break;
                }

                ret = listen(sock, 10);
                if (ret != SUCCESS) {
                    ERROR("listen() - %s\n", strerror(errno));
                    break;
                }

                svc_if[i].sock = sock;

                INFO("TinyDHT IPv6 service listening on port %hu\n", 
                        TINYDHT_SERVICE);

                break;

            default:
                break;
        }
    }

    return SUCCESS;
}

int
tinydht_add_poll_fd(int fd)
{
    if (n_poll_fds >= MAX_POLL_FD) {
        return FAILURE;
    }

    poll_fd[n_poll_fds] = fd;
    n_poll_fds++;

    return SUCCESS;
}

int
tinydht_add_task(struct task *task)
{
    TAILQ_INSERT_TAIL(&task_list, task, next);
    return SUCCESS;
}

struct dht *
tinydht_find_dht_from_fd(int fd)
{
    int i;

    for (i = 0; i < n_dhts; i++) {
        if ((dht[i]->net_if.sock == fd)) {
            return dht[i];
        }
    }

    return NULL;
}

bool
tinydht_is_service_fd(int fd)
{
    int i;

    for (i = 0; i < n_svc_ifs; i++) {
        if (svc_if[i].sock == fd) {
            return TRUE;
        }
    }

    return FALSE;
}

int
tinydht_poll_loop(void)
{
    struct pollfd fds[MAX_POLL_FD];
    int i;
    int ret;
    bool data_avail;
    bool fd_found;
    struct dht *dht = NULL;
    u8 buf[2048];
    struct sockaddr_storage from;
    size_t fromlen;
    int len = 0;
    int sock;

    INFO("TinyDHT polling %d fds\n", n_poll_fds);

    bzero(fds, sizeof(fds));

    for (i = 0; i < n_poll_fds; i++) {
        fds[i].fd = poll_fd[i];
        fds[i].events = POLLIN | POLLERR | POLLHUP | POLLNVAL;
    }

    while (TRUE) {

        /* call the task_scheduler */
        tinydht_task_schedule();
        
        data_avail = FALSE;
        errno = 0;

        ret = poll(fds, n_poll_fds, MAX_POLL_TIMEOUT);

        switch (ret) {
            case -1:        /* error */
                return FAILURE;
            case 0:         /* timeout */
                continue;
            default:        /* some data is ready */
                ASSERT(ret == 1);
                data_avail = TRUE;
                break;
        }


        if (!data_avail) {
            continue;
        }

        fd_found = FALSE;
        for (i = 0; i < n_poll_fds; i++) {
            if (fds[i].revents & POLLIN) {
                fd_found = TRUE;
                break;
            }
        }

        if (data_avail && !fd_found) {
            return FAILURE;
        }

        if (tinydht_is_service_fd(poll_fd[i])) {

            sock = accept(poll_fd[i], (struct sockaddr *)&from, &fromlen);
            if (ret < 0) {
                ERROR("accept() - %s\n", strerror(errno));
                continue;
            }

            len = recv(sock, buf, sizeof(buf), 0);
            if (len <= 0) {
                ERROR("recv() - %s\n", strerror(errno));
                continue;
            }

            ret = tinydht_decode_request(sock, &from, fromlen, buf, len);
            if (ret < 0) {
                continue;
            }

        } else {

            /* read the data */
            len = recvfrom(poll_fd[i], buf, sizeof(buf), 0, 
                    (struct sockaddr *)&from, &fromlen);
            if (len <= 0) {
                ERROR("recvfrom() - %s\n", strerror(errno));
                continue;
            }

            /* has it arrived on a dht instance? */
            dht = tinydht_find_dht_from_fd(poll_fd[i]);
            if (!dht) {
                return FAILURE;
            }

            ret = dht->decode_rpc(dht, &from, fromlen, buf, len);
            if (ret != SUCCESS) {
                continue;
            }
        }
    }

    return FAILURE;
}

int
tinydht_task_schedule(void)
{
    struct task *task = NULL;
    bool found = FALSE;
    int ret;

    TAILQ_FOREACH(task, &task_list, next) {

        found = FALSE;

        switch (task->state) {
            case TASK_STATE_RUNNABLE:
                found = TRUE;
                break;
            case TASK_STATE_WAIT:
                break;
            default:
                break;
        }

        if (found) {
            break;
        }
    }

    if (found) {
        ret = task->dht->task_schedule(task);
        if (ret != SUCCESS) {
            return ret;
        }
    }

    return SUCCESS;
}

int
tinydht_decode_request(int sock, struct sockaddr_storage *from, 
                        size_t fromlen, u8 *data, int len)
{
    struct tinydht_msg_req req;
    struct tinydht_msg_rsp rsp;
    struct tinydht_msg *msg = NULL;
    int ret;

    bzero(&req, sizeof(req));
    memcpy(&req, data, len);
    req.key_len = htonl(req.key_len);
    req.val_len = htonl(req.val_len);

    msg = (struct tinydht_msg *) malloc(sizeof(struct tinydht_msg));
    if (!msg) {
        ERROR("%s\n", strerror(errno));
        goto err;
    }

    bzero(msg, sizeof(struct tinydht_msg));
    memcpy(&msg->req, &req, sizeof(req));
    msg->sock = sock;
    memcpy(&msg->from, from, fromlen);
    msg->fromlen = fromlen;

    switch (req.action) {
        case TINYDHT_ACTION_PUT:
            INFO("PUT received\n");
            ret = tinydht_put(msg);
            if (ret != SUCCESS) {
                goto err;
            }

            break;

        case TINYDHT_ACTION_GET:
            INFO("GET received\n");
            ret = tinydht_get(msg);
            if (ret != SUCCESS) {
                goto err;
            }

            break;

        default:
            ERROR("unknown action %d\n", req.action);
            free(msg);
            goto err;
    }

    return SUCCESS;

err:

    bzero(&rsp, sizeof(rsp));
    rsp.status = TINYDHT_RESPONSE_FAILURE;

    ret = send(sock, &rsp, sizeof(rsp), 0);
    if (ret < 0) {
        ERROR("send() - %s\n", strerror(errno));
    }

    return FAILURE;
}

int
tinydht_put(struct tinydht_msg *msg)
{
    int i;
    bool found = FALSE;
    int ret;

    for (i = 0; i < n_dhts; i++) {
        ret = dht[i]->put(dht[i], msg);
        if (ret != SUCCESS) {
            continue;
        }
        found = TRUE;
    }

    if (found) {
        return SUCCESS;
    } 

    return FAILURE;
}

int
tinydht_get(struct tinydht_msg *msg)
{
    int i;
    bool found = FALSE;
    int ret;

    for (i = 0; i < n_dhts; i++) {
        ret = dht[i]->get(dht[i], msg);
        if (ret != SUCCESS) {
            continue;
        }
        found = TRUE;
    }

    if (found) {
        return SUCCESS;
    } 

    return FAILURE;
}
